import numba as nb
import numpy as np
from scipy.stats import rankdata,norm
from scipy.sparse import issparse
from tqdm import tqdm
from liana.method.sp._utils import _zscore, _spatialdm_weight_norm


class LocalFunction:
    """
    Class representing information about bivariate spatial functions.
    """

    instances = {}

    def __init__(self, name, metadata, fun, reference=None):
        self.name = name
        self.metadata = metadata
        self.fun = fun
        self.reference = reference

        LocalFunction.instances[name] = self

    def __call__(self,
                 x_mat,
                 y_mat,
                 weight,
                 n_perms,
                 seed,
                 mask_negatives,
                 verbose
                 ):
        if self.name == 'morans':
            x_mat = self._norm_max(x_mat)
            y_mat = self._norm_max(y_mat)
            weight = _spatialdm_weight_norm(weight)
        else:
            if issparse(x_mat):
               x_mat = x_mat.A
            if issparse(y_mat):
                y_mat = y_mat.A

        if self.name.__contains__("masked") or weight.shape[0] < 10000:
            weight = weight.todense().A

        local_scores = self.fun(x_mat, y_mat, weight)

        if n_perms is None:
            local_pvals = None
        elif n_perms > 0:
            local_pvals = self._permutation_pvals(x_mat=x_mat,
                                                  y_mat=y_mat,
                                                  weight=weight,
                                                  local_truth=local_scores,
                                                  n_perms=n_perms,
                                                  seed=seed,
                                                  mask_negatives=mask_negatives,
                                                  verbose=verbose
                                                  )
        elif n_perms == 0:
            local_pvals = self._zscore_pvals(x_mat=x_mat,
                                             y_mat=y_mat,
                                             weight=weight,
                                             local_truth=local_scores,
                                             mask_negatives=mask_negatives
                                             )

        return local_scores, local_pvals

    def __repr__(self):
        return f"{self.name}: {self.metadata}"

    def _permutation_pvals(self,
                           x_mat,
                           y_mat,
                           weight,
                           local_truth,
                           n_perms,
                           seed,
                           mask_negatives,
                           verbose):
        rng = np.random.default_rng(seed)

        spot_n = local_truth.shape[0]
        xy_n = local_truth.shape[1]

        local_pvals = np.zeros((spot_n, xy_n))

        # shuffle the matrix
        for i in tqdm(range(n_perms), disable=not verbose):
            _idx = rng.permutation(spot_n)
            perm_score = self.fun(x_mat=x_mat[_idx, :], y_mat=y_mat[_idx, :], weight=weight)
            if mask_negatives:
                local_pvals += np.array(perm_score >= local_truth, dtype=int)
            else:
                local_pvals += np.array(np.abs(perm_score) >= np.abs(local_truth), dtype=int)

        local_pvals = local_pvals / n_perms

        return local_pvals


    def _zscore_pvals(self, x_mat, y_mat, local_truth, weight, mask_negatives):
        """
        Local Moran's R analytical p-values as in spatialDM (Li et al., 2022)


        Parameters
        ----------
        x_mat
            2D array with x variables
        y_mat
            2D array with y variables
        local_r
            2D array with Local Moran's I
        weight
            connectivity weights
        mask_negatives
            Whether to mask negative correlations pvalue

        Returns
        -------
        2D array of p-values with shape(n_spot, xy_n)

        """
        spot_n = x_mat.shape[0]

        x_norm = np.apply_along_axis(norm.fit, axis=0, arr=x_mat)
        y_norm = np.apply_along_axis(norm.fit, axis=0, arr=y_mat)

        # get x,y std
        x_sigma, y_sigma = x_norm[1, :], y_norm[1, :]

        x_sigma = x_sigma * spot_n / (spot_n - 1)
        y_sigma = y_sigma * spot_n / (spot_n - 1)

        std = self._get_local_var(x_sigma, y_sigma, weight, spot_n)
        local_zscores = local_truth / std

        if mask_negatives:
            local_zpvals = norm.sf(local_zscores)
        else:
            local_zpvals = norm.sf(np.abs(local_zscores))

        return local_zpvals


    def _get_local_var(self, x_sigma, y_sigma, weight, spot_n):
        """
        Spatial weight variance as in spatialDM (Li et al., 2022)

        Parameters
        ----------
        x_sigma
            Standard deviations for each x (e.g. std of all ligands in the matrix)
        y_sigma
            Standard deviations for each y (e.g. std of all receptors in the matrix)
        weight
            connectivity weight matrix
        spot_n
            number of spots/cells in the matrix

        Returns
        -------
        2D array of standard deviations with shape(n_spot, xy_n)

        """
        if not isinstance(weight, np.ndarray):
            weight = np.array(weight.todense())

        weight_sq = (weight ** 2).sum(axis=1)

        dim = 2 * (spot_n - 1) ** 2 / spot_n ** 2
        sigma_prod = x_sigma * y_sigma
        core = dim * sigma_prod

        var = np.multiply.outer(weight_sq, core) + core
        std = var ** 0.5

        return std

    def _norm_max(self, X, axis=0):
        X = X / X.max(axis=axis).A
        X = _zscore(X, axis=axis)
        X = np.where(np.isnan(X), 0, X)

        return X

    @classmethod
    def _get_instance(cls, name):
        name = name.lower()
        instances = cls.instances
        if name not in instances:
            raise ValueError(f"Function {name} not found. Available functions are: {', '.join(instances)}")

        return cls.instances[name]


@nb.njit(nb.float32(nb.float32[:], nb.float32[:], nb.float32[:], nb.float32), cache=True)
def _wcorr(x, y, w, wsum):

    x = np.argsort(x).argsort().astype(nb.float32)
    y = np.argsort(y).argsort().astype(nb.float32)

    wx = w * x
    wy = w * y

    numerator = wsum * sum(wx * y) - sum(wx) * sum(wy)

    denominator_x = wsum * sum(w * (x**2)) - sum(wx)**2
    denominator_y = wsum * sum(w * (y**2)) - sum(wy)**2
    denominator = (denominator_x * denominator_y)

    if (denominator == 0) or (numerator == 0):
        return 0

    return numerator / (denominator**0.5)


@nb.njit(nb.float32[:,:](nb.float32[:,:], nb.float32[:,:], nb.float32[:,:]), parallel=True, cache=True)
def _masked_spearman(x_mat, y_mat, weight):
    spot_n = x_mat.shape[0]
    xy_n = x_mat.shape[1]

    local_corrs = np.zeros((spot_n, xy_n), dtype=nb.float32)

    for i in nb.prange(spot_n):
        w = weight[i, :]
        msk = w > 0
        wsum = sum(w[msk])

        for j in range(xy_n):
            x = x_mat[:, j][msk]
            y = y_mat[:, j][msk]

            local_corrs[i, j] = _wcorr(x, y, w[msk], wsum)

    # NOTE done due to numpy/numba sum imprecision, https://github.com/numba/numba/issues/8749
    local_corrs = np.clip(a=local_corrs, a_min=-1.0, a_max=1.0, out=local_corrs)

    return local_corrs


def _vectorized_correlations(x_mat, y_mat, weight, method="pearson"):
    """
    Vectorized implementation of weighted correlations.

    Note: due to the imprecision of np.sum and np.dot, the function is accurate to 5 decimal places.

    """
    if method not in ["pearson", "spearman"]:
        raise ValueError("method must be one of 'pearson', 'spearman'")
    weight_sums = np.array(np.sum(weight, axis=1)).reshape(-1, 1)

    if method=="spearman":
        x_mat = rankdata(x_mat, axis=0)
        y_mat = rankdata(y_mat, axis=0)

    # standard pearson
    n1 = weight_sums * (weight @ (x_mat * y_mat))
    n2 = (weight @ x_mat) * (weight @ y_mat)
    numerator = n1 - n2

    denominator_x = (weight_sums * (weight @ x_mat ** 2)) - (weight @ x_mat)**2
    denominator_y = (weight_sums * (weight @ y_mat ** 2)) - (weight @ y_mat)**2
    denominator = denominator_x * denominator_y

    # numpy sum is unstable below 1e-6...
    denominator[denominator < 1e-6] = 0
    denominator = denominator ** 0.5

    zeros = np.zeros(numerator.shape)
    local_corrs = np.divide(numerator, denominator, out=zeros, where=denominator!=0)

    # NOTE done due to numpy/numba sum imprecision, https://github.com/numba/numba/issues/8749
    local_corrs = np.clip(local_corrs, -1, 1, out=local_corrs, dtype=np.float32)

    return local_corrs


def _vectorized_pearson(x_mat, y_mat, weight):
    return _vectorized_correlations(x_mat, y_mat, weight, method="pearson")


def _vectorized_spearman(x_mat, y_mat, weight):
    return _vectorized_correlations(x_mat, y_mat, weight, method="spearman")


def _vectorized_cosine(x_mat, y_mat, weight):
    xy_dot = weight @ (x_mat * y_mat)
    x_dot = weight @ (x_mat**2)
    y_dot = weight @ (y_mat**2)
    denominator = (x_dot * y_dot) + np.finfo(np.float32).eps

    return xy_dot / denominator**0.5


def _vectorized_jaccard(x_mat, y_mat, weight):
    x_mat, y_mat = x_mat > 0, y_mat > 0 ## NOTE only positive
    numerator = weight @ np.minimum(x_mat, y_mat)
    denominator = weight @ np.maximum(x_mat, y_mat) + np.finfo(np.float32).eps

    return numerator / denominator


def _local_morans(x_mat, y_mat, weight):
    """

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables

    Returns
    -------
    Returns 2D array of local Moran's I with shape(n_spot, xy_n)

    """
    local_x = x_mat * (weight @ y_mat)
    local_y = y_mat * (weight @ x_mat)
    local_r = (local_x + local_y)

    return local_r


def _product(x_mat, y_mat, weight):
    x_mat = weight @ x_mat
    y_mat = weight @ y_mat
    score = x_mat * y_mat

    return score


def _norm_product(x_mat, y_mat, weight):
    x_mat = weight @ x_mat
    y_mat = weight @ y_mat

    x_norm = np.max(np.abs(x_mat), axis=0)
    y_norm = np.max(np.abs(y_mat), axis=0)

    x_norm[x_norm == 0.] = 1.
    y_norm[y_norm == 0.] = 1.

    x_mat = x_mat / x_norm
    y_mat = y_mat / y_norm

    score = x_mat * y_mat

    return score

_bivariate_functions = [
        LocalFunction(
            name="pearson",
            metadata="weighted Pearson correlation coefficient",
            fun = _vectorized_pearson,
        ),
        LocalFunction(
            name="spearman",
            metadata="weighted Spearman correlation coefficient",
            fun = _vectorized_spearman,
        ),
        LocalFunction(
            name="cosine",
            metadata="weighted Cosine similarity",
            fun = _vectorized_cosine,
        ),
        LocalFunction(
            name="jaccard",
            metadata="weighted Jaccard similarity",
            fun = _vectorized_jaccard,
        ),
        LocalFunction(
            name="product",
            metadata="simple weighted product",
            fun = _product,
            reference="If vars are z-scaled = Lee's static (Lee 2021;J.Geograph.Syst.)"
        ),
        LocalFunction(
            name="norm_product",
            metadata="normalized weighted product",
            fun = _norm_product,
        ),
        LocalFunction(
            name="morans",
            metadata="Moran's R",
            fun=_local_morans,
            reference="Li, Z., Wang, T., Liu, P. and Huang, Y., 2022. SpatialDM:"
            "Rapid identification of spatially co-expressed ligand-receptor"
            "reveals cell-cell communication patterns. bioRxiv, pp.2022-08."
        ),
        LocalFunction(
            name= "masked_spearman",
            metadata="masked & weighted Spearman correlation",
            fun=_masked_spearman,
            reference="Ghazanfar, S., Lin, Y., Su, X., Lin, D.M., Patrick, E., Han, Z.G., Marioni, J.C. and Yang, J.Y.H., 2020."
            "Investigating higher-order interactions in single-cell data with scHOT. Nature methods, 17(8), pp.799-806."
        ),
    ]



import pandas as pd
import numpy as np
from liana._logging import _logg, _check_if_installed
from liana._constants import DefaultValues as V


def build_prior_network(ppis,
                        input_nodes,
                        output_nodes,
                        lr_sep=None,
                        verbose=V.verbose):
    """
    Build Prior Network from PPIs and input/output nodes.

    Parameters
    ----------
    ppis : list of tuples or pandas DataFrame
        The PPIs to use for the prior network. If a pandas DataFrame is provided, it must have the columns
    input_nodes : dict
        A dictionary of input nodes. The keys are the node names, the values are the node scores.
    output_nodes : dict
        A dictionary of output nodes. The keys are the node names, the values are the node scores.
    lr_sep : str, optional
        The separator to use to split the input nodes into ligand and receptor. If None, the input nodes will be used as is.
    verbose : bool, optional
        Whether to print progress information. Default: True

    Returns
    -------
    corneto.Graph

    """

    cn = _check_if_installed("corneto")

    if lr_sep is not None:
        if any(lr_sep in k for k in input_nodes.keys()):
            _logg(f"Input nodes are in the format Ligand{lr_sep}Receptor. Extracting only the receptor...", verbose=verbose)
            # Print only at most the first 3 entries
            for k, v in list(input_nodes.items())[:3]:
                _logg(f" - {k} -> {v}", verbose=verbose)
            _logg(" - ...", verbose=verbose)
            # Do the split only if lr_sep is present, otherwise get the same key:
            input_nodes = {k.split(lr_sep)[-1]: v for k, v in input_nodes.items()}

    _logg("Importing network...", verbose=verbose)
    if isinstance(ppis, list):
        G = cn.Graph.from_sif_tuples(ppis)
    elif isinstance(ppis, pd.DataFrame):
        ppis = [(r.source, r.mor, r.target) for (_, r) in ppis.iterrows() if r.source != r.target]
        G = cn.Graph.from_sif_tuples(ppis)
    else:
        raise ValueError("PPIs must be a list of tuples or a pandas DataFrame.")
    _logg("done.", verbose=verbose)

    incl_inputs = set(G.vertices).intersection(set(input_nodes))
    incl_outputs = set(G.vertices).intersection(set(output_nodes))

    _logg(f" - Nodes x Edges: {G.num_vertices, G.num_edges}", verbose=verbose)
    _logg(f" - Provided inputs included in the prior network: {len(incl_inputs)}/{len(input_nodes)}", verbose=verbose)
    _logg(f" - Provided outputs included in the network: {len(incl_outputs)}/{len(output_nodes)}", verbose=verbose)

    _logg("Performing reachability analysis...", verbose=verbose)
    Gp = G.prune(list(incl_inputs), list(incl_outputs))
    _logg("done.", verbose=verbose)
    pruned_size = (Gp.num_vertices, Gp.num_edges)
    incl_inputs_pruned = set(Gp.vertices).intersection(incl_inputs)
    incl_outputs_pruned = set(Gp.vertices).intersection(incl_outputs)

    _logg(f" - Selected inputs: {len(incl_inputs_pruned)}/{len(incl_inputs)}.", verbose=verbose)
    _logg(f" - Selected outputs: {len(incl_outputs_pruned)}/{len(incl_outputs)}.", verbose=verbose)
    _logg(f" - Final size of the prior graph: {pruned_size}.", verbose=verbose)

    if len(incl_outputs_pruned) == 0:
        raise ValueError("None of the output nodes can be reached from the provided input nodes in the PPI network.")
    if len(incl_inputs_pruned) == 0:
        raise ValueError("None of the input nodes can reach any of the output nodes in the PPI network.")

    return Gp

def _get_scores(d):
    return (
       [v for v in d.values() if v < 0],
       [v for v in d.values() if v > 0]
    )


def find_causalnet(
        prior_graph,
        input_node_scores,
        output_node_scores,
        node_weights=None,
        node_cutoff=0.1,
        min_penalty=0.01,
        max_penalty=1.0,
        missing_penalty=10,
        edge_penalty=0.01,
        solver=None,
        seed=1337,
        max_runs=1,
        stable_runs=5,
        verbose=True,
        **kwargs
        ):
    """
    Find the causal network that best explains the input/output node scores.

    Parameters
    ----------
    prior_graph : corneto.Graph
        The prior graph to use for the search.
    input_node_scores : dict
        A dictionary of input node scores.
    output_node_scores : dict
        A dictionary of output node scores.
    node_weights : dict, optional
        A dictionary of node weights. The keys are the node names, the values are the weights.
        If None, all nodes will have the same weight.
    node_cutoff : float
        The cutoff to use for the node weights. Nodes with a weight below this cutoff will be assigned
        the max_penalty, nodes with a weight above this cutoff will be assigned the min_penalty.
        Only used if node_weights is not None. Default: 0.1
    min_penalty : float
        The minimum penalty to assign to nodes with a weight above the cutoff.
        Only used if node_weights is not None. Default: 0.01
    max_penalty : float
        The maximum penalty to assign to nodes with a weight below the cutoff
        Only used if node_weights is not None. Default: 1.0
    missing_penalty : float
        The penalty to assign to nodes that are not measured. Default: 10
    edge_penalty : float
        The penalty to assign to edges. Default: 0.01
    solver : str, optional
        The solver to use. If None, the default solver will be used. Default: None
        It will default to the solver included in SCIPY, if no other solver is available.
    seed : int, optional
        The seed to use for the random number generator. Default: 1337
    max_runs : int, optional
        The maximum number of runs to perform. Consider increasing this value if the solver does not converge.
        In each run, the noise added to the edge and node penalties is perturbed slightly (iterating over the seed).
        By default, only 1 run is performed.
    stable_runs : int, optional
        The number of consecutive stable solutions requires to interrupt the iteration over max_runs. Only used if max_runs is not == 1. Default: 5
    verbose : bool, optional
        Whether to print progress information. Default: True
    **kwargs : dict, optional
        Additional arguments to pass to the solver.
    """

    cn = _check_if_installed("corneto")

    if solver is None:
        solver = cn.methods.carnival.select_mip_solver()

    measured_nodes = set(input_node_scores.keys()) | set(output_node_scores.keys())

    _logg("Total positive/negative scores of the inputs and outputs:", verbose=verbose)
    w_neg_in, w_pos_in = _get_scores(input_node_scores)
    w_neg_out, w_pos_out = _get_scores(output_node_scores)
    _logg(f" - (-) input nodes: {sum(w_neg_in)}", verbose=verbose)
    _logg(f" - (+) input nodes: {sum(w_pos_in)}", verbose=verbose)
    _logg(f" - (-) output nodes: {sum(w_neg_out)}", verbose=verbose)
    _logg(f" - (+) output nodes: {sum(w_pos_out)}", verbose=verbose)

    # Total weights
    total = abs(sum(w_neg_in)) + abs(sum(w_neg_out)) + sum(w_pos_in) + sum(w_pos_out)
    _logg(f" - abs total (inputs + outputs): {total}", verbose=verbose)

    if node_weights is None:
        node_penalties = {}
    else:
        node_penalties = _weights_to_penalties(node_weights,
                                               cutoff=node_cutoff,
                                               max_penalty=max_penalty,
                                               min_penalty=min_penalty)

    run_count = 0 # total runs
    stable_count = 0 # stable solutions in a row
    df_all = None # df with all solutions

    while run_count < max_runs:
        current_seed = seed + run_count
        if run_count > 0:
            _logg(f"Run {run_count} with seed {current_seed}", verbose=verbose)

        # assign 0 penalties to input/output nodes, missing_penalty to missing nodes
        # add a small amount of noise to the penalties to ensure reproducible solutions
        rng = np.random.default_rng(seed=current_seed)
        c_node_penalties = {k: node_penalties.get(k, missing_penalty) + rng.uniform(min_penalty/20, min_penalty/10)
                            if k not in measured_nodes else 0.0 for k in prior_graph.vertices}

        _logg("Building CORNETO problem...", verbose=verbose)
        P, G = cn.methods.carnival._extended_carnival_problem(
            prior_graph,
            input_node_scores,
            output_node_scores,
            node_penalties=c_node_penalties,
            edge_penalty=edge_penalty
        )

        # E is the variable with 1 if edge activates or inhibits, 0 otherwise
        E = P.symbols['reaction_sends_activation_c0'] + P.symbols['reaction_sends_inhibition_c0']
        W = rng.uniform(edge_penalty / 20, edge_penalty / 10, size=E.shape)
        P.add_objectives(W.T @ E)

        _logg(f"Solving with {solver}...", verbose=verbose)
        if (solver=='scipy') and verbose:
            kwargs.update(scipy_options=dict(disp='true'))

        P.solve(
            solver=solver,
            verbosity=int(verbose),
            **kwargs)

        obj_names = ["Loss (unfitted inputs/output)", "Edge penalty error", "Node penalty error"]
        _logg("Solution summary:", verbose=verbose)
        for s, o in zip(obj_names, P.objectives):
            _logg(f" - {s}: {o.value}", verbose=verbose)

        rows, cols = cn.methods.carnival.export_results(P, G, input_node_scores, output_node_scores)
        df = pd.DataFrame(rows, columns=cols)

        # Check if all rows from df are contained in df_all
        if df_all is None:
            df_all = df
            continue
        else:
            set_df = set(tuple(row) for row in df.values)
            set_df_all = set(tuple(row) for row in df_all.values)

            if set_df.issubset(set_df_all):
                stable_count += 1
            else:
                stable_count = 0

            df_all = pd.concat([df_all, df]).drop_duplicates()

        if stable_count >= stable_runs:
            break

        run_count += 1

    return df_all, P


def _weights_to_penalties(props,
                          cutoff,
                          min_penalty,
                          max_penalty):
    if any(p < 0 or p > 1 for p in props.values()):
        raise ValueError("Node weights were not between 0 and 1. Consider minmax or another normalization.")

    return {k: max_penalty if v < cutoff else min_penalty for k, v in props.items()}



from liana._constants import PrimaryColumns as P, DefaultValues as V
from numpy import union1d


def _join_stats(source, target, dedict, resource):
    """
    Joins and renames source-ligand and target-receptor stats to the ligand-receptor resource

    Parameters
    ----------
    source
        Source/Sender cell type
    target
        Target/Receiver cell type
    dedict
        dictionary
    resource
        Ligand-receptor Resource

    Returns
    -------
    Ligand-Receptor stats

    """
    source_stats = dedict[source].copy()
    source_stats.columns = source_stats.columns.map(
        lambda x: 'ligand_' + str(x))
    source_stats = source_stats.rename(
        columns={'ligand_names': 'ligand', 'ligand_label': 'source'})

    target_stats = dedict[target].copy()
    target_stats.columns = target_stats.columns.map(
        lambda x: 'receptor_' + str(x))
    target_stats = target_stats.rename(
        columns={'receptor_names': 'receptor', 'receptor_label': 'target'})

    bound = resource.merge(source_stats).merge(target_stats)

    return bound

def _get_props(X_mask):
    return X_mask.getnnz(axis=0) / X_mask.shape[0]

def _get_groupby_subset(groupby_pairs):
    if groupby_pairs is not V.groupby_pairs:
        if not (P.source in groupby_pairs.columns) | (P.target in groupby_pairs.columns):
            raise AssertionError(f"{P.source} and {P.target} must be in groupby_pairs")
        groupby_subset = union1d(groupby_pairs[P.source].unique(),
                                 groupby_pairs[P.target].unique())

    else:
        groupby_subset = None
    return groupby_subset



import numpy as np

from liana.method.sp._bivariate._local_functions import _vectorized_pearson, _vectorized_spearman, \
    _vectorized_cosine, _vectorized_jaccard, _masked_spearman, _local_morans, _product, \
    _norm_product

from scipy.sparse import csr_matrix


rng = np.random.default_rng(seed=0)

x_mat = rng.normal(size=(20, 5)).astype(np.float32)
y_mat = rng.normal(size=(20, 5)).astype(np.float32)
weight = csr_matrix(rng.uniform(size=(20, 20)).astype(np.float32))

def _assert_bivariate(function, desired, x_mat, y_mat, weight):
    actual = function(x_mat, y_mat, weight)
    assert actual.shape == (20, 5)
    np.testing.assert_almost_equal(actual[0, :], desired, decimal=5)


def test_pc_vectorized():
    pc_vec_truth = np.array([0.25005114, 0.04262733, -0.00130362, 0.2903336, -0.1236529])
    _assert_bivariate(_vectorized_pearson, pc_vec_truth, x_mat, y_mat, weight)


def test_sp_vectorized():
    sp_vec_truth = np.array([0.23636213, 0.16480759, -0.01487235, 0.22840601, -0.11492937])
    _assert_bivariate(_vectorized_spearman, sp_vec_truth, x_mat, y_mat, weight)


def test_sp_masked():
    sp_masked_truth = np.array([0.23636216, 0.16480756, -0.0148723, 0.22840606, -0.11492944])
    _assert_bivariate(_masked_spearman, sp_masked_truth, x_mat, y_mat, weight.A)  # NOTE the .A is to convert to dense


def test_costine_vectorized():
    cosine_vec_truth = np.array([0.33806977, 0.03215113, 0.0950243, 0.2957758, -0.10259595])
    _assert_bivariate(_vectorized_cosine, cosine_vec_truth, x_mat, y_mat, weight)


def test_vectorized_jaccard():
    jaccard_vec_truth = np.array([0.34295967, 0.35367563, 0.39685577, 0.41780996, 0.30527356])
    _assert_bivariate(_vectorized_jaccard, jaccard_vec_truth, x_mat, y_mat, weight)


# NOTE: spatialdm uses raw counts
def test_morans():
    sp_morans_truth = np.array([-1.54256, 0.64591, 1.30025, 0.55437, -0.77182])
    _assert_bivariate(_local_morans, sp_morans_truth, x_mat, y_mat, weight)


def test_product():
    product_vec_truth = np.array([5.4518123, -0.7268728, 8.350364, 0.53861964, 1.4466602])
    _assert_bivariate(_product, product_vec_truth, x_mat, y_mat, weight.A)


def test_norm_product():
    product_vec_truth = np.array([0.4081537, -0.03988646, 0.42921585, 0.03255661, 0.08895018])
    _assert_bivariate(_norm_product, product_vec_truth, x_mat, y_mat, weight.A)



from __future__ import annotations

import anndata
import pandas

from plotnine import ggplot, geom_point, aes, \
    facet_grid, labs, theme_bw, theme, element_text, element_rect, scale_size_continuous, scale_color_cmap

from liana.plotting._common import _prep_liana_res, _check_var, _get_top_n, _filter_by, _inverse_scores

from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V

@d.dedent
def dotplot(adata: anndata.AnnData = None,
            uns_key = K.uns_key,
            liana_res: pandas.DataFrame = None,
            colour: str = None,
            size: str = None,
            source_labels: list = None,
            target_labels: list = None,
            top_n: int = None,
            orderby: str | None = None,
            orderby_ascending: bool | None = None,
            orderby_absolute: bool = False,
            filter_fun: callable = None,
            ligand_complex: str | None = None,
            receptor_complex: str | None = None,
            inverse_colour: bool = False,
            inverse_size: bool = False,
            cmap: str = V.cmap,
            size_range: tuple = (2, 9),
            figure_size: tuple = (8, 6),
            return_fig=V.return_fig
            ) -> ggplot:
    """
    Dotplot interactions by source and target cells

    Parameters
    ----------
    %(adata)s
    %(uns_key)s
    %(liana_res)s
    %(colour)s
    %(size)s
    %(source_labels)s
    %(target_labels)s
    %(top_n)s
    %(orderby)s
    %(orderby_ascending)s
    %(orderby_absolute)s
    %(filter_fun)s
    %(ligand_complex)s
    %(receptor_complex)s
    %(inverse_colour)s
    %(inverse_size)s
    %(cmap)s
    %(size_range)s
    %(figure_size)s

    Returns
    -------
    A `plotnine.ggplot` instance

    """
    liana_res = _prep_liana_res(adata=adata,
                                liana_res=liana_res,
                                source_labels=source_labels,
                                target_labels=target_labels,
                                ligand_complex = ligand_complex,
                                receptor_complex = receptor_complex,
                                uns_key=uns_key
                                )
    _check_var(liana_res, var=colour, var_name='colour')
    _check_var(liana_res, var=size, var_name='size')

    liana_res = _filter_by(liana_res, filter_fun)
    liana_res = _get_top_n(liana_res, top_n, orderby, orderby_ascending, orderby_absolute)

    # inverse sc if needed
    if inverse_colour:
        liana_res[colour] = _inverse_scores(liana_res[colour])
    if inverse_size:
        liana_res[size] = _inverse_scores(liana_res[size])

    # generate plot
    p = (ggplot(liana_res, aes(x='target', y='interaction', colour=colour, size=size))
         + geom_point()
         + facet_grid('~source')
         + scale_size_continuous(range=size_range)
         + scale_color_cmap(cmap)
         + labs(color=str.capitalize(colour),
                size=str.capitalize(size),
                y="Interactions (Ligand -> Receptor)",
                x="Target",
                title="Source")
         + theme_bw()
         + theme(legend_text=element_text(size=14),
                 strip_background=element_rect(fill="white"),
                 strip_text=element_text(size=15, colour="black"),
                 axis_text_y=element_text(size=10, colour="black"),
                 axis_title_y=element_text(colour="#808080", face="bold", size=15),
                 axis_text_x=element_text(size=11, face="bold", angle=90),
                 figure_size=figure_size,
                 plot_title=element_text(vjust=0, hjust=0.5, face="bold",
                                         colour="#808080", size=15)
                 )
         )

    if return_fig:
        return p

    p.draw()


@d.dedent
def dotplot_by_sample(adata: anndata.AnnData  = None,
                      uns_key: str = K.uns_key,
                      liana_res: pandas.DataFrame = None,
                      sample_key: str = 'sample',
                      colour: str  = None,
                      size: str = None,
                      inverse_colour: bool = False,
                      inverse_size: bool = False,
                      source_labels: str | None = None,
                      target_labels: str | None = None,
                      ligand_complex: str | None = None,
                      receptor_complex: str | None = None,
                      size_range: tuple = (2, 9),
                      cmap: str = V.cmap,
                      figure_size: tuple = (8, 6),
                      return_fig: bool = V.return_fig
                      ):
    """
    A dotplot of interactions by sample

    Parameters
    ----------
        %(adata)s
        %(uns_key)s
        %(liana_res)s
        sample_key
            sample_key used to group different samples/contexts from `liana_res`. Defaults to 'sample'.
        %(colour)s
        %(size)s
        %(inverse_colour)s
        %(inverse_size)s
        %(source_labels)s
        %(target_labels)s
        %(ligand_complex)s
        %(receptor_complex)s
        %(size_range)s
        %(cmap)s
        %(figure_size)s

    Returns
    -------
    Returns a ggplot for the specified interactions by sample.

    """

    liana_res = _prep_liana_res(adata=adata,
                                liana_res=liana_res,
                                source_labels=source_labels,
                                target_labels=target_labels,
                                ligand_complex=ligand_complex,
                                receptor_complex=receptor_complex,
                                uns_key=uns_key)
    _check_var(liana_res, var=colour, var_name='colour')
    _check_var(liana_res, var=size, var_name='size')

    # inverse sc if needed
    if inverse_colour:
        liana_res[colour] = _inverse_scores(liana_res[colour])
    if inverse_size:
        liana_res[size] = _inverse_scores(liana_res[size])

    p = (ggplot(liana_res, aes(x='target', y='source', colour=colour, size=size))
            + geom_point()
            + facet_grid(f'interaction~{sample_key}', space='free', scales='free')
            + scale_size_continuous(range=size_range)
            + scale_color_cmap(name=cmap)
            + labs(color=str.capitalize(colour),
                   size=str.capitalize(size),
                   y="Source",
                   x="Target",
                   title=sample_key.capitalize())
            + theme_bw()
            + theme(legend_text=element_text(size=14),
                    strip_background=element_rect(fill="white"),
                    strip_text=element_text(size=13, colour="black", angle=90),
                    axis_text_y=element_text(size=10, colour="black"),
                    axis_title_y=element_text(colour="#808080", face="bold", size=12),
                    axis_text_x=element_text(size=11, face="bold", angle=90),
                    axis_title_x=element_text(colour="#808080", face="bold", size=12),
                    figure_size=figure_size,
                    plot_title=element_text(vjust=0, hjust=0.5, face="bold", size=12),
                    )
            )
    if return_fig:
        return p

    p.draw()



import numpy as np
from sklearn.neighbors import BallTree
from plotnine import ggplot, aes, geom_line, geom_point, theme_bw, xlab, ylab
from pandas import DataFrame

def query_bandwidth(coordinates: np.ndarray,
                    start: int = 0,
                    end: int = 500,
                    interval_n: int = 50,
                    reference: np.ndarray = None
                    ):
    """
    Query the bandwidth (maximum distance) at which the average number of neighbors is maximized.

    Parameters
    ----------
    coordinates
        Spatial coordinates of spots.
    start
        Starting bandwidth.
    end
        Ending bandwidth.
    interval_n
        Number of bandwidths to query. Used to generate a linearly spaced interval between `start` and `end`.
        Default is 50.
    reference
        Reference coordinates to query the neighbors from. Default is `None`, which will use `coordinates`.

    Returns
    -------
    A `plotnine` plot and a `pandas` DataFrame with the following columns:
        - `bandwith`: the bandwidth (maximum distance) at which the average number of neighbors is maximized.
        - `neighbours`: the average number of neighbors at the specified bandwidth.
    """
    tree = BallTree(coordinates, metric='euclidean')
    df = DataFrame()
    interval = np.linspace(start, end, interval_n)

    if reference is None:
        _reference = coordinates
    else:
        _reference = reference

    for n in range(interval_n):
        max_distance = interval[n]
        df.loc[n, 'bandwith'] = max_distance

        # query the neighbors within the specified distance
        num_neighbors = tree.query_radius(_reference, r=max_distance, count_only=True)

        # calculate the average number of neighbors
        avg_nn = np.ceil(np.median(num_neighbors))
        df.loc[n, 'neighbours'] = avg_nn - 1

    p = (ggplot(df, aes(x='bandwith', y='neighbours')) +
         geom_line() +
         geom_point() +
         theme_bw(base_size=16) +
         xlab("Bandwidth") +
         ylab("Number of Neighbors")
         )

    return p, df



from itertools import product
import urllib.request
import os

import numpy as np
import pandas as pd
from liana._logging import _logg

def _replace_subunits(lst, my_dict, one_to_many):
    result = []
    for x in lst:
        if x in my_dict:
            value = my_dict[x]

            if not isinstance(value, list):
                value = [value]

            if len(value) > one_to_many:
                result.append(np.nan)
            else:
                result.append(value)
        else:
            result.append(np.nan)
    return result


def _generate_orthologs(data, column, map_dict, one_to_many):
    df = data[[column]].drop_duplicates().set_index(column)

    df["subunits"] = df.index.str.split("_")
    df["subunits"] = df["subunits"].apply(
        _replace_subunits,
        args=(
            map_dict,
            one_to_many,
        ),
    )
    df = df["subunits"].explode().reset_index()

    grouped = (
        df.groupby(column).filter(lambda x: x["subunits"].notna().all()).groupby(column)
    )

    # Generate all possible subunit combinations within each group
    complexes = []
    for name, group in grouped:
        if group["subunits"].isnull().all():
            continue
        subunit_lists = [list(x) for x in group["subunits"]]
        complex_combinations = list(product(*subunit_lists))
        for complex in complex_combinations:
            complexes.append((name, "_".join(complex)))

    # Create output DataFrame
    col_names = ["orthology_source", "orthology_target"]
    result = pd.DataFrame(complexes, columns=col_names).set_index("orthology_source")

    return result


def translate_column(
    resource,
    map_df,
    column,
    replace=True,
    one_to_many=1,
    ):
    """
    Generate orthologs for a given column in a DataFrame.

    Parameters
    ----------
    resource : pandas.DataFrame
        Input DataFrame.
    map_df : pandas.DataFrame
        DataFrame with orthology mappings, where the first column is the source and the second column is the target for mapping.
    column : str
        Column name to translate.
    replace : bool, optional
        Whether to replace the original column with the translated values. Default is True.
        If False, it will create a new column with the prefix "orthology_".
    one_to_many : int, optional
        Maximum number of orthologs allowed per gene. Default is 1.

    Details
    -------
    This function generates orthologs for a given column in a DataFrame.
    It handles complex names by splitting them into subunits and generating all possible combinations of orthologs.
    It assumes that subunits are separated by an underscore ("_").

    Returns
    -------
    Resulting DataFrame with translated column.

    """
    if not isinstance(one_to_many, int):
        raise ValueError("`one_to_many` should be a positive integer!")
    if ['source', 'target'] != map_df.columns.tolist():
        raise ValueError("The `map_df` DataFrame must have two columns named 'source' and 'target'!")

    # get orthologs
    map_df = map_df.set_index("source")
    map_dict = map_df.groupby(level=0)["target"].apply(list).to_dict()
    map_data = _generate_orthologs(resource, column, map_dict, one_to_many)

    # join orthologs
    resource = resource.merge(map_data,
                              left_on=column,
                              right_index=True,
                              how="left")

    # replace orthologs
    if replace:
        resource[column] = resource["orthology_target"]
    else:
        resource[f"orthology_{column}"] = resource.apply(
            lambda x: x["orthology_target"]
            if not pd.isnull(x["orthology_target"])
            else x[column],
            axis=1,
        )
    resource = resource.drop(columns=["orthology_target"])

    resource = resource.dropna(subset=[column])
    return resource


# function that loops over columns and applies translate_column
def translate_resource(resource, map_df, columns=['ligand', 'receptor'], **kwargs):
    """
    Generate orthologs for multiple columns in a DataFrame.

    Parameters
    ----------
    resource : pandas.DataFrame
        Input DataFrame.
    map_df : pandas.DataFrame
        DataFrame with orthology mappings, where the first column is the source and the second column is the target for mapping.
    columns : list
        List of column names to translate.
    **kwargs
        Additional arguments for `liana.utils.translate_column`.

    Returns
    -------
    Resulting DataFrame with translated columns.

    """
    for column in columns:
        resource = translate_column(resource, map_df, column, **kwargs)

    return resource


def get_hcop_orthologs(url="https://ftp.ebi.ac.uk/pub/databases/genenames/hcop/human_mouse_hcop_fifteen_column.txt.gz",
                       filename=None,
                       min_evidence=3,
                       columns = None
                       ):
    """
    Simple function to download the HCOP file from the EBI FTP server and filter it by minimum evidence.

    Parameters
     ----------
    url : str
        URL of the HCOP file. See https://ftp.ebi.ac.uk/pub/databases/genenames/hcop/ for bulk download options besides human and mouse.
    filename : str
        Name of the file to save the HCOP file.
    min_evidence : int
        Minimum number of evidences to keep the interaction, where evidence is the number of orthology resources supporting the interaction.
    columns : list
        Columns to keep in the final DataFrame. If None, it will keep the default columns.

    Returns
    -------
    mapping : pd.DataFrame
        DataFrame with the HCOP mapping.

    Details
    -------
    HCOP is a composite database combining data from various orthology resources.
    It provides a comprehensive set of orthologs among human, mouse, and rat, among many other species.

    If you use this function, please reference the original HCOP papers:
    - Eyre, T.A., Wright, M.W., Lush, M.J. and Bruford, E.A., 2007. HCOP: a searchable database of human orthology predictions. Briefings in bioinformatics, 8(1), pp.2-5.
    - Yates, B., Gray, K.A., Jones, T.E. and Bruford, E.A., 2021. Updates to HCOP: the HGNC comparison of orthology predictions tool. Briefings in Bioinformatics, 22(6), p.bbab155.

    For more information, please visit the HCOP website: https://www.genenames.org/tools/hcop/,
    or alternatively check the bulk download FTP links page: https://ftp.ebi.ac.uk/pub/databases/genenames/hcop/
    """
    # check if exists
    if filename is None:
        filename = os.path.basename(url.split("/")[-1])
    if not os.path.exists(filename):
        urllib.request.urlretrieve(url, filename)
    else:
        _logg(f"File {filename} already exists. Skipping download.", level="info")

    mapping = pd.read_csv(filename, sep="\t")
    mapping['evidence'] = mapping['support'].apply(lambda x: len(x.split(",")))
    mapping = mapping[mapping['evidence'] >= min_evidence]

    if columns is not None:
        mapping = mapping[columns]

    return mapping



from __future__ import annotations

import numpy as np
from scipy.stats import norm
from tqdm import tqdm
from liana._logging import _logg
from liana.method.sp._utils import _zscore, _spatialdm_weight_norm

class GlobalFunction:
    instances = {}

    def __init__(self, fun, name):
        self.fun = fun
        self.name = name
        self.pvals_name = self.name+ '_pvals'

        GlobalFunction.instances[name] = self

    def _permutation_pvals(self,
                           x_mat,
                           y_mat,
                           weight,
                           global_stat,
                           n_perms,
                           mask_negatives,
                           seed,
                           verbose
                           ):
        rng = np.random.default_rng(seed)

        # initialize mat /w n_perms * number of X->Y
        idx = x_mat.shape[0]

        # permutation mat /w n_permss x LR_n
        perm_mat = np.zeros((n_perms, global_stat.shape[0]))

        for perm in tqdm(range(n_perms), disable=not verbose):
            _idx = rng.permutation(idx)
            perm_mat[perm, :] = self.fun(x_mat=x_mat[_idx, :],
                                         y_mat=y_mat[_idx, :],
                                         weight=weight)

        if mask_negatives:
            global_pvals = 1 - (global_stat > perm_mat).sum(axis=0) / n_perms
        else:
            global_pvals = 1 - (np.abs(global_stat) > np.abs(perm_mat)).sum(axis=0) / n_perms

        return global_pvals


    def _zscore_pvals(self,
                      weight,
                      global_stat,
                      mask_negatives
                      ):
        """
        SpatialDM's global z-score p-value calculation

        """
        if not isinstance(weight, np.ndarray):
            weight = np.array(weight.todense())
        spot_n = weight.shape[0]

        # global distance/weight variance as in spatialDM
        numerator = spot_n ** 2 * ((weight * weight).sum()) - \
                    (2 * spot_n * (weight @ weight).sum()) + \
                    (weight.sum() ** 2)
        denominator = spot_n ** 2 * (spot_n - 1) ** 2
        weight_var_sq = (numerator / denominator) ** (1 / 2)

        global_zscores = global_stat / weight_var_sq

        if mask_negatives:
            global_zpvals = norm.sf(global_zscores)
        else:
            global_zpvals = norm.sf(np.abs(global_zscores)) * 2

        return global_zpvals


    def __call__(self,
                 xy_stats,
                 x_mat,
                 y_mat,
                 weight,
                 seed,
                 n_perms,
                 mask_negatives,
                 verbose
                 ):
        if self.name == 'morans':
            x_mat = _zscore(x_mat, axis=0, global_r=True)
            y_mat = _zscore(y_mat, axis=0, global_r=True)
            weight = _spatialdm_weight_norm(weight)
        elif self.name == 'lee':
            x_mat = _zscore(x_mat)
            y_mat = _zscore(y_mat)
            weight = weight * weight
        else:
            raise ValueError('Global function not supported')

        global_stat = self.fun(x_mat=x_mat, y_mat=y_mat, weight=weight)

        if n_perms is None:
            global_pvals = None
        elif n_perms > 0:
            global_pvals = \
                self._permutation_pvals(x_mat=x_mat,
                                        y_mat=y_mat,
                                        weight=weight,
                                        global_stat=global_stat,
                                        n_perms=n_perms,
                                        mask_negatives=mask_negatives,
                                        seed=seed,
                                        verbose=verbose
                                        )
        elif n_perms==0 and self.name == 'morans':
            global_pvals = \
                self._zscore_pvals(weight=weight,
                                   global_stat=global_stat,
                                   mask_negatives=mask_negatives
                                   )
        elif n_perms==0 and self.name == 'lee':
            global_pvals = None
            _logg('Global Lee does not support analytical p-values', 'warning', verbose=verbose)

        xy_stats[self.name] = global_stat
        xy_stats[self.pvals_name] = global_pvals


def _global_r(x_mat, y_mat, weight):
    return ((weight @ x_mat) * y_mat).sum(axis=0)


def _global_l(x_mat, y_mat, weight):
    return ((weight @ x_mat) * y_mat).sum(axis=0) / weight.sum()

_global_r = GlobalFunction(_global_r, 'morans')
_global_l = GlobalFunction(_global_l, 'lee')



from liana.method.sp._bivariate._spatial_bivariate import bivariate
from liana.method.sp._misty._misty_constructs import lrMistyData, MistyData, genericMistyData
from liana.method.sp._misty._single_view_models import SingleViewModel, RandomForestModel, LinearModel, RobustLinearModel



import logging
from types import ModuleType

logging.basicConfig(level=logging.INFO, format='%(message)s')

def _logg(message, level='info', verbose=False):
    """
    Log a message with a specified logging level.

    Args:
        message (str): The message to log.
        level (int or None): The logging level for the message (default is None).
            If None or False, no logging is performed.
    """
    if verbose:
        if level == "warn":
            logging.warning(message)
        elif level == "info":
            logging.info(message)

def _check_if_installed(package_name: str, custom_error_message: str = None) -> ModuleType:
    try:
        imported_module = __import__(package_name)
        return imported_module
    except ImportError:
        if custom_error_message:
            raise ImportError(custom_error_message)
        else:
            raise ImportError(f'{package_name} is not installed. Please install it with: pip install {package_name}')



from liana.multi import to_tensor_c2c, adata_to_views, lrs_to_views, filter_view_markers
from liana.utils._getters import get_factor_scores, get_variable_loadings
from liana.testing import sample_lrs

import numpy as np
import pandas as pd

import cell2cell as c2c

from liana.testing._sample_anndata import generate_toy_adata

adata = generate_toy_adata()


def test_to_tensor_c2c():
    """Test to_tensor_c2c."""
    liana_res = sample_lrs(by_sample=True)

    liana_dict = to_tensor_c2c(liana_res=liana_res,
                               sample_key='sample',
                               score_key='specificity_rank',
                               return_dict=True
                               )
    assert isinstance(liana_dict, dict)

    tensor = to_tensor_c2c(liana_res=liana_res,
                           sample_key='sample',
                           score_key='specificity_rank')
    assert isinstance(tensor, c2c.tensor.tensor.PreBuiltTensor)
    assert tensor.sparsity_fraction()==0.0


def test_lrs_to_views():
    """Test lrs_to_views."""
    liana_res = sample_lrs(by_sample=True)
    adata.uns['liana_results'] = liana_res

    mdata = lrs_to_views(adata=adata,
                         sample_key='sample',
                         score_key='specificity_rank',
                         uns_key = 'liana_results',
                         obs_keys = ['case'],
                         source_key='source',
                         target_key='target',
                         ligand_key='ligand_complex',
                         receptor_key='receptor_complex',
                         lr_prop=0.1,
                         lrs_per_sample=0,
                         lrs_per_view=5,
                         samples_per_view=0,
                         min_variance=-1, # don't filter
                         verbose=True
                         )

    assert mdata.shape == (4, 16)
    assert 'case' in mdata.obs.columns
    assert len(mdata.varm_keys())==3



def test_adata_to_views():
    """Test adata_to_views."""
    mdata = adata_to_views(adata,
                           groupby='bulk_labels',
                           sample_key='sample',
                           obs_keys=None,
                           min_cells=5,
                           min_counts=10,
                           keep_stats=False,
                           mode='sum',
                           verbose=True,
                           use_raw=True,
                           min_smpls=2,
                           # filter features
                           min_count=0,
                           min_total_count=0,
                           large_n=0,
                           min_prop=0,
                           skip_checks=True # skip because it's log-normalized
                           )

    assert len(mdata.varm_keys())==9
    assert 'case' not in mdata.obs.columns
    assert mdata.shape == (4, 6201)
    assert 'psbulk_stats' not in mdata.uns.keys()

    # test feature level filtering (with default values)
    mdata = adata_to_views(adata,
                           groupby='bulk_labels',
                           sample_key='sample',
                           obs_keys = ['case'],
                           mode='sum',
                           keep_stats=True,
                           verbose=True,
                           use_raw=True,
                           skip_checks=True
                           )

    assert len(mdata.varm_keys())==7
    assert 'case' in mdata.obs.columns
    assert mdata.shape == (4, 1598)
    assert mdata.uns['psbulk_stats'] is not None


def test_filter_view_markers():
    mdata = adata_to_views(adata,
                           groupby='bulk_labels',
                           sample_key='sample',
                           obs_keys = ['case'],
                           mode='sum',
                           verbose=True,
                           use_raw=True,
                           skip_checks=True
                           )

    rng = np.random.default_rng(42)
    markers = {}
    for cell_type in mdata.mod.keys():
        markers[cell_type] = rng.choice(adata.var_names, 10).tolist()

    filter_view_markers(mdata, markers, inplace=True)
    assert mdata.mod['Dendritic'].var['highly_variable'].sum() == 139

    filter_view_markers(mdata, markers, var_column=None, inplace=True)
    assert mdata.shape == (4, 1471)


def test_get_funs():
    liana_res = sample_lrs(by_sample=True)
    adata.uns['liana_results'] = liana_res

    mdata = lrs_to_views(adata=adata,
                         sample_key='sample',
                         score_key='specificity_rank',
                         uns_key = 'liana_results',
                         lr_prop=0.1,
                         lrs_per_sample=0,
                         lrs_per_view=5,
                         samples_per_view=0,
                         min_variance=-1, # don't filter
                         verbose=True
                         )

    # generate random loadings
    mdata.varm['LFs'] = np.random.rand(mdata.shape[1], 5)

    loadings = get_variable_loadings(mdata,
                                     varm_key='LFs',
                                     view_sep=':',
                                     variable_sep='^',
                                     pair_sep='&')
    assert isinstance(loadings, pd.DataFrame)
    assert loadings.shape == (16, 9)

    # dont drop columns & and don't separate
    loadings = get_variable_loadings(mdata,
                                     varm_key='LFs',
                                     drop_columns=False)
    assert isinstance(loadings, pd.DataFrame)
    assert loadings.shape == (16, 6)

    # generate random factor scores
    mdata.obsm['X_mofa'] = np.random.rand(mdata.shape[0], 5)

    scores = get_factor_scores(mdata, obsm_key='X_mofa')
    assert isinstance(scores, pd.DataFrame)
    assert scores.shape == (4, 6)




from pandas import Series
import numpy as np
from scipy.sparse import csr_matrix, isspmatrix_csr, hstack
from anndata import AnnData

def _add_complexes_to_var(adata, entities, complex_sep='_'):
    """
    Generate an AnnData object with complexes appended as variables.
    """
    complexes = entities[Series(entities).str.contains(complex_sep)]
    X = None
    for comp in complexes:
        subunits = comp.split(complex_sep)

        # keep only complexes, the subunits of which are in var
        if all([subunit in adata.var.index for subunit in subunits]):
            adata.var.loc[comp, :] = None

            # create matrix for this complex
            new_array = csr_matrix(adata[:, subunits].X.min(axis=1))

            if X is None:
                X = new_array
            else:
                X = hstack((X, new_array))

    adata = AnnData(X=hstack((adata.X, X)),
                    obs=adata.obs,
                    var=adata.var,
                    obsm=adata.obsm,
                    varm = adata.varm,
                    obsp=adata.obsp,
                    uns = adata.uns,
                    )

    if not isspmatrix_csr(adata.X):
        adata.X = adata.X.tocsr()

    return adata


def _zscore(mat, axis=0, global_r=False):
    if global_r: # NOTE: specific to global SpatialDM
        spot_n = 1
    else:
        spot_n = mat.shape[axis]

    mat = mat - mat.mean(axis=axis)
    mat = mat / np.sqrt(np.sum(np.power(mat, 2), axis=axis) / spot_n)
    mat = np.clip(mat, -10, 10)

    return np.array(mat)

def _spatialdm_weight_norm(weight):
    norm_factor = weight.shape[0] / weight.sum()
    weight = norm_factor * weight
    return weight



from mudata import MuData
import numpy as np
from pandas import concat
from liana.utils import obsm_to_adata
from liana.method._pipe_utils import prep_check_adata
from liana._constants import DefaultValues as V
from liana._logging import _check_if_installed

def estimate_metalinks(adata,
                       resource,
                       pd_net,
                       t_net=None,
                       x_name='metabolite',
                       y_name='receptor',
                       use_raw=V.use_raw,
                       layer=V.layer,
                       verbose=V.verbose,
                       **kwargs):
    """
    Estimate Metabolites from anndata object, and return a MuData object of metabolites and receptors.

    Parameters
    ----------
    adata
        Annotated data matrix.
    resource
        Resource to use for ligand-receptor inference.
    pd_net
        Metabolic production-degradation network to use.
    t_net
        Transport set to use.
    x_name
        Name of the metabolite modality.
    y_name
        Name of the receptor modality. Must be present as a column in the resource.
    **kwargs
        Additional arguments to pass to the decoupler-py functions.
        Method-specific arguments are not supported.

    Returns
    -------
    A MuData object with metabolite & receptor assays.

    """
    dc = _check_if_installed(package_name="decoupler")
    ad = prep_check_adata(adata,
                          layer=layer,
                          use_raw=use_raw,
                          verbose=verbose,
                          groupby=None,
                          min_cells=None,
                          uns=adata.uns,
                          obsm=adata.obsm
                          )
    dc.run_ulm(ad, net = pd_net, use_raw=False, verbose=verbose, **kwargs)
    met_est = ad.obsm['ulm_estimate']

    if t_net is not None:
        dc.run_wmean(ad, t_net, times=0, use_raw=False, verbose=verbose, **kwargs)

        out_est = ad.obsm['wmean_estimate']
        intersect = np.intersect1d(met_est.columns, out_est.columns)
        out_est = out_est[intersect]

        out_mask = out_est > 0

        mask = np.ones(out_est.shape)
        mask[out_mask == 0] = 0

        # mask those with transporters
        mmat = met_est[intersect] * mask

        # concat the rest
        coldiff = np.setdiff1d(met_est.columns, mmat.columns)
        mmat = concat([mmat, met_est[coldiff]], axis = 1)

    else:
        mmat = met_est

    resource = resource[resource[x_name].isin(mmat.columns)].copy()
    receptor = ad[:, ad.var.index.isin(np.unique(resource[y_name]))]

    ad.obsm['mmat'] = mmat
    mmat = obsm_to_adata(ad, 'mmat')

    return MuData({x_name:mmat, y_name:receptor})



from scipy.sparse import csr_matrix, isspmatrix_csr
import numpy as np

def zi_minmax(X, cutoff=0.5):
    """
    Zero-inflated min-max scaling, adopted from CiteFuse (Kim et al., 2020; https://academic.oup.com/bioinformatics/article/36/14/4137/5827474).

    Parameters
    ----------
    X : array-like
        Data to be scaled.
    cutoff : float
        Cutoff value for zero-inflation - values less than this are set to 0. Default is 0.5.

    Returns
    -------
    X : csr_matrix

    """

    X = X.copy()
    if not isspmatrix_csr(X):
        X = csr_matrix(X)

    min_vals = np.array(X.min(axis=0).todense())[0]
    max_vals = np.array(X.max(axis=0).todense())[0]
    nonzero_rows, nonzero_cols = X.nonzero()
    scaled_values = (X.data - min_vals[nonzero_cols]) / (max_vals[nonzero_cols] - min_vals[nonzero_cols])

    scaled_values[scaled_values < cutoff] = 0
    nonzero_rows, nonzero_cols = X.nonzero()

    X = csr_matrix((scaled_values, (nonzero_rows, nonzero_cols)), shape=X.shape)

    return X

def neg_to_zero(X, cutoff=0):
    """
    Set negative values to 0.

    Parameters
    ----------
    X : array-like
        Data to be transformed.
    cutoff : float
        Cutoff value for zero-inflation - values less than this are set to 0. Default is 0.

    Returns
    -------
    A csr_matrix.

    """
    X = X.copy()
    if not isspmatrix_csr(X):
        X = csr_matrix(X)
    X.data[X.data < cutoff] = 0
    return X



import numpy as np
import pandas as pd
from numpy import random
from scanpy.datasets import pbmc68k_reduced
from liana.utils.spatial_neighbors import spatial_neighbors

def generate_toy_spatial():
    adata = pbmc68k_reduced()

    rng = np.random.default_rng(seed=1337)
    x = rng.integers(low=0, high=5000, size=adata.shape[0])
    y = rng.integers(low=0, high=5000, size=adata.shape[0])
    adata.obsm['spatial'] = np.array([x, y]).T
    spatial_neighbors(adata, bandwidth=100, cutoff=0.1, set_diag=True)

    return adata

def generate_toy_mdata():
    import scanpy as sc
    from mudata import MuData

    adata = generate_toy_spatial()
    adata = adata.raw.to_adata()
    adata = adata[:, 0:10]
    sc.pp.filter_cells(adata, min_counts=1)

    adata.layers['scaled'] = sc.pp.scale(adata.X, zero_center=True, max_value=5)

    adata_y = adata.copy()

    # create mdata
    mdata = MuData({'adata_x':adata, 'adata_y':adata_y})
    mdata.obsp = adata.obsp
    mdata.uns = adata.uns
    mdata.obsm = adata.obsm
    mdata.obs = adata.obs

    return mdata


def generate_toy_adata():
    adata = pbmc68k_reduced()
    sample_key = 'sample'

    rng = random.default_rng(0)

    # create fake samples
    adata.obs[sample_key] = rng.choice(['A', 'B', 'C', 'D'], size=len(adata.obs))

    # group samples into conditions
    adata.obs['case'] = adata.obs[sample_key].map({'A': 'yes', 'B': 'yes', 'C': 'no', 'D': 'no'})

    return adata


def generate_anndata(sparsity = 0.90, n_ct = 10, n_vars = 2000, n_obs = 1000, seed=1337):
    # TODO, eventually change completely to use this function, inplace of the other ones
    from scipy.sparse import csr_matrix
    import scanpy as sc
    from liana.utils import spatial_neighbors

    rng = np.random.default_rng(seed=seed)
    counts = rng.poisson(100, size=(n_obs, n_vars))
    mask = rng.choice([0, 1], size=(n_obs, n_vars), p=[sparsity, 1 - sparsity])
    counts = csr_matrix(counts * mask, dtype=np.float32)

    adata = sc.AnnData(counts)
    sc.pp.filter_cells(adata, min_counts=1)
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata)

    adata.var_names = [f"Gene{i:d}" for i in range(adata.n_vars)]
    adata.obs_names = [f"Cell{i:d}" for i in range(adata.n_obs)]
    print(f" NNZ fraction: {adata.X.nnz / (adata.X.shape[0] * adata.X.shape[1])}")

    x = rng.integers(low=0, high=5000, size=adata.shape[0])
    y = rng.integers(low=0, high=5000, size=adata.shape[0])
    adata.obsm['spatial'] = np.array([x, y]).T

    spatial_neighbors(adata, cutoff=0.1, bandwidth=150, max_neighbours=10)

    # assign cell types
    ct = rng.choice([f"CT{i:d}" for i in range(n_ct)], size=(adata.n_obs,))
    ct = rng.choice(ct, size=(adata.n_obs,))
    adata.obs["cell_type"] = pd.Categorical(ct)

    return adata



from itertools import product
import pandas as pd

def sample_resource(adata, n_lrs = 3000, seed=1337):
    resource = pd.DataFrame(product(adata.var_names, adata.var_names)).rename(columns={0: "ligand", 1: "receptor"})
    resource = resource[resource["ligand"] != resource["receptor"]]
    resource = resource.sample(n_lrs, replace=False, random_state=seed)
    return resource



from liana.resource.select_resource import _handle_resource, select_resource
import pytest

def test_select_interactions():
    # generate list of 2x strings
    interactions = [('a', 'b'), ('c', 'd')]

    resource = _handle_resource(interactions=interactions,
                                x_name='x',
                                y_name='y',
                                verbose=True,
                                # These should be ignored
                                resource=select_resource("consensus"),
                                resource_name='consensus')

    assert resource.shape[0] == 2
    assert (resource.columns == ['x', 'y']).all()


def test_select_resource():
    resource = _handle_resource(interactions=None,
                                x_name='ligand',
                                y_name='receptor',
                                verbose=True,
                                resource=select_resource("consensus"),
                                # This should be ignored
                                resource_name='ignore me'
                                )

    assert resource.shape[0] == 4624
    assert (resource.columns == ['ligand', 'receptor']).all()


def test_select_resource_name():
    resource = _handle_resource(interactions=None,
                                resource=None,
                                resource_name='cellchatdb',
                                x_name='x',
                                y_name='y',
                                verbose=True,
                                )

    assert resource.shape[0] == 1912
    assert (resource.columns == ['ligand', 'receptor']).all()


def test_resource_exception_none():
    with pytest.raises(ValueError):
        _handle_resource(interactions=None,
                         resource=None,
                         resource_name=None,
                         x_name='x',
                         y_name='y',
                         verbose=True,)


def test_resource_exception_xy():
    with pytest.raises(ValueError):
        _handle_resource(interactions=None,
                         resource=select_resource("consensus"),
                         resource_name=None,
                         x_name='x',
                         y_name='y',
                         verbose=True,
                         )



import pandas as pd
import numpy as np

def _sample_interactions():
    columns = ['target', 'predictor', 'view', 'importances']
    interactions = pd.DataFrame(columns=columns)
    interactions['target'] = np.repeat(['a', 'b', 'c'], 3)
    interactions['predictor'] = np.tile(['x', 'y', 'z'], 3)
    interactions['view'] = np.repeat(['intra', 'inter', 'extra'], 3)
    interactions['importances'] = np.random.rand(9)

    return interactions


def _sample_target_metrics():
    columns = ['target', 'intra_R2', 'multi_R2', 'gain_R2', 'intra', 'extra']
    target_metrics = pd.DataFrame(columns=columns)
    target_metrics['target'] = ['a', 'b', 'c']
    target_metrics['intra_R2'] = np.random.rand(3)
    target_metrics['multi_R2'] = np.random.rand(3)
    target_metrics['gain_R2'] = np.random.rand(3)
    target_metrics['intra'] = np.random.rand(3)
    target_metrics['extra'] = np.random.rand(3)

    return target_metrics



from anndata import AnnData
import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import normalize

from liana._docs import d
from liana._constants import Keys as K, DefaultValues as V

def _gaussian(distance_mtx, bandwidth):
    return np.exp(-(distance_mtx ** 2.0) / (2.0 * bandwidth ** 2.0))

def _misty_rbf(distance_mtx, bandwidth):
    return np.exp(-(distance_mtx ** 2.0) / (bandwidth ** 2.0))

def _exponential(distance_mtx, bandwidth):
    return np.exp(-distance_mtx / bandwidth)

def _linear(distance_mtx, bandwidth):
    connectivity = 1 - distance_mtx / bandwidth
    return np.clip(connectivity, a_min=0, a_max=np.inf)


@d.dedent
def spatial_neighbors(adata: AnnData,
                      bandwidth=None,
                      cutoff=0.1,
                      max_neighbours=None,
                      kernel='gaussian',
                      set_diag=False,
                      zoi=0,
                      standardize=False,
                      reference=None,
                      spatial_key=K.spatial_key,
                      key_added=K.spatial_key,
                      inplace=V.inplace
                      ):
    """
    Generate spatial connectivity weights using Euclidean distance.

    Parameters
    ----------

    %(adata)s
    bandwidth
         Denotes signaling length (`l`) and controls the maximum distance at which two spots are considered.
         Corresponds to the units in which spatial coordinates are expressed.
    cutoff
        Values below this cutoff will be set to 0.
    max_neighbours
        Maximum nearest neighbours to be considered when generating spatial connectivity weights.
        Essentially, the maximum number of edges in the graph. Default is `None`, which will use n = adata.shape[0]/10.
    kernel
        Kernel function used to generate connectivity weights.
        It controls the shape of the connectivity weights.
        The following options are available: ['gaussian', 'exponential', 'linear', 'misty_rbf']
    set_diag
        Logical, sets connectivity diagonal to 0 if `False`. Default is `True`.
    zoi
        Zone of indifference. Values below this cutoff will be set to `np.inf`.
    standardize
        Whether to (l1) standardize spatial proximities (connectivities) so that they sum to 1.
        This plays a role when weighing border regions prior to downstream methods, as the number of spots
        in the border region (and hence the sum of proximities) is smaller than the number of spots in the center.
        Relevant for methods with unstandardized scores (e.g. product). Default is `False`.
    reference
        Reference coordinates to use when generating spatial connectivity weights.
        If `None`, uses the spatial coordinates in `adata.obsm[spatial_key]`.
        This is only relevant if you want to use a different set of coordinates to generate spatial connectivity weights.
    %(spatial_key)s
    key_added
        Key to add to `adata.obsp` if `inplace = True`. If reference is not `None`, key will be added to `adata.obsm`.
    %(inplace)s

    Notes
    -----
    This function is adapted from mistyR, and is set to be consistent with
    the `squidpy.gr.spatial_neighbors` function in the `squidpy` package.

    Returns
    -------
    If ``inplace = False``, returns an `np.array` with spatial connectivity weights.
    Otherwise, modifies the ``adata`` object with the following key:
        - :attr:`anndata.AnnData.obsp` ``['{key_added}_connectivities']`` with the aforementioned array

    """

    if cutoff is None:
        raise ValueError("`cutoff` must be provided!")
    assert spatial_key in adata.obsm
    families = ['gaussian', 'exponential', 'linear', 'misty_rbf']
    if kernel not in families:
        raise AssertionError(f"{kernel} must be a member of {families}")
    if bandwidth is None:
        raise ValueError("Please specify a bandwidth")

    coordinates = adata.obsm[spatial_key]

    if reference is None:
        _reference = coordinates
    else:
        _reference = reference

    if max_neighbours is None:
        max_neighbours = int(adata.shape[0] / 10)

    tree = NearestNeighbors(n_neighbors=max_neighbours + 1, # +1 to exclude self
                            algorithm='ball_tree',
                            metric='euclidean').fit(_reference)
    dist = tree.kneighbors_graph(coordinates, mode='distance')

    # prevent float overflow
    bandwidth = np.array(bandwidth, dtype=np.float64)

    # define zone of indifference
    dist.data[dist.data < zoi] = np.inf

    # NOTE: dist gets converted to a connectivity (proximity) matrix
    if kernel == 'gaussian':
        dist.data = _gaussian(dist.data, bandwidth)
    elif kernel == 'misty_rbf':
        dist.data = _misty_rbf(dist.data, bandwidth)
    elif kernel == 'exponential':
        dist.data = _exponential(dist.data, bandwidth)
    elif kernel == 'linear':
        dist.data = _linear(dist.data, bandwidth)
    else:
        raise ValueError("Please specify a valid family to generate connectivity weights")

    if not set_diag:
        dist.setdiag(0)
    if cutoff is not None:
        dist.data = dist.data * (dist.data > cutoff)
    if standardize:
        dist = normalize(dist, axis=1, norm='l1')

    spot_n = dist.shape[0]
    if reference is None:
        assert spot_n == adata.shape[0]
    if spot_n > 1000:
        dist = dist.astype(np.float32)

    if inplace:
        if reference is not None:
            adata.obsm[f'{key_added}_connectivities'] = dist
        else:
            adata.obsp[f'{key_added}_connectivities'] = dist

    return None if inplace else dist



import pytest
from anndata import AnnData
import numpy as np
from liana.utils.interpolate_adata import interpolate_adata

def create_test_adata(n_cells, n_genes, spatial_key='spatial'):
    """
    Helper function to create a test AnnData object.
    """
    X = np.random.rand(n_cells, n_genes)
    obsm = {spatial_key: np.random.rand(n_cells, 2)}
    adata = AnnData(X, obsm=obsm)
    adata.layers['some_layer'] = X
    return adata

@pytest.fixture
def reference_adata():
    return create_test_adata(100, 10)

@pytest.fixture
def target_adata():
    return create_test_adata(80, 10)

def test_basic_interpolation(reference_adata, target_adata):
    result = interpolate_adata(reference=reference_adata, target=target_adata, spatial_key='spatial', use_raw=False)
    assert isinstance(result, AnnData)
    assert result.shape == (100, 10)

@pytest.mark.parametrize("method", ['linear', 'nearest'])
def test_different_methods(reference_adata, target_adata, method):
    result = interpolate_adata(reference=reference_adata, target=target_adata, spatial_key='spatial', method=method, use_raw=False)
    assert isinstance(result, AnnData)

def test_fill_value(reference_adata, target_adata):
    fill_value = -1
    result = interpolate_adata(reference=reference_adata, target=target_adata, spatial_key='spatial', fill_value=fill_value, use_raw=False)
    assert (result.X.data == fill_value).sum() > 0

def test_invalid_spatial_key(reference_adata, target_adata):
    with pytest.raises(KeyError):
        interpolate_adata(reference=reference_adata, target=target_adata, spatial_key='invalid_key')

def test_use_raw_layer_parameters(reference_adata, target_adata):
    result_layer = interpolate_adata(reference=reference_adata, target=target_adata, spatial_key='spatial', layer='some_layer', use_raw=False)
    assert isinstance(result_layer, AnnData)

    with pytest.raises(ValueError):
        interpolate_adata(reference=reference_adata, target=target_adata, spatial_key='spatial', layer='some_layer', use_raw=True)



import numpy as np

from liana.method.sc._Method import Method, MethodMeta, _show_methods
from liana.method.sc._rank_aggregate import AggregateClass, _rank_aggregate_meta as aggregate_meta
from liana.method.sc import cellphonedb, connectome, logfc, natmi, singlecellsignalr, geometric_mean, cellchat, scseqcomm

from liana.method.sp import bivariate, genericMistyData, lrMistyData, MistyData
from liana.method.fun._causalnet import find_causalnet, build_prior_network
from liana.method.fun._estimate_metalinks import estimate_metalinks
from liana._constants import DefaultValues as V

# callable consensus instance
_methods = [cellphonedb, connectome, logfc, natmi, singlecellsignalr]
rank_aggregate = AggregateClass(aggregate_meta, methods=_methods)


def show_methods():
    """Shows methods available in LIANA"""
    return _show_methods(_methods + [rank_aggregate, geometric_mean, scseqcomm, cellchat])

def get_method_scores():
    """Returns a dict of all scoring functions, with a boolean indicating whether the score is ascending or not"""
    instances = np.array(MethodMeta.instances)
    relevant = np.array([(isinstance(instance, Method)) | (isinstance(instance, AggregateClass)) for instance in instances])
    instances = instances[relevant]

    specificity_scores = {method.specificity: method.specificity_ascending for method in instances if method.specificity is not None}
    magnitude_scores = {method.magnitude : method.magnitude_ascending for method in instances if method.magnitude is not None}

    scores = {**specificity_scores, **magnitude_scores}
    return scores

def process_scores(liana_res, score_key, inverse_fun=V.inverse_fun):

    df = liana_res.copy()
    scores = get_method_scores()

    if not np.isin(score_key, list(scores.keys())).any():
        raise ValueError(f"Score column {score_key} not found in liana's method scores.")

    # reverse if ascending order
    ascending_order = scores[score_key]
    if(ascending_order):
        df[score_key] = inverse_fun(df[score_key])

    return df



from scipy.stats import gmean

from liana.method.sc._Method import Method, MethodMeta
from liana.method._pipe_utils._get_mean_perms import _calculate_pvals

def _gmean_score(x, perm_stats) -> tuple:
    """
    Calculate CellPhoneDB-like LR means and p-values

    Parameters
    ----------
    x
        DataFrame with LIANA results
    perm_stats
        Permutation statistics with shape (2 (ligand-receptor), n_perms (number of permutations), n_rows (in lr_res)

    Returns
    -------
    A tuple with lr_mean and p-value for x

    """
    lr_gmeans = gmean((x['ligand_means'].values, x['receptor_means'].values), axis=0)
    gmean_pvals = _calculate_pvals(lr_gmeans, perm_stats, gmean)

    return lr_gmeans, gmean_pvals

_geometric_mean = MethodMeta(method_name="Geometric Mean",
                             complex_cols=["ligand_means", "receptor_means"],
                             add_cols=[],
                             fun=_gmean_score,
                             magnitude="lr_gmeans",
                             magnitude_ascending=False,
                             specificity="gmean_pvals",
                             specificity_ascending=True,
                             permute=True,
                             reference="CellPhoneDBv2's permutation approach applied to the "
                                       "geometric means of ligand-receptors' mean, as opposed to "
                                       "their arithmetic mean."
                             )

geometric_mean = Method(_method=_geometric_mean)



import numpy as np
import pandas as pd
from anndata import AnnData

from types import ModuleType
from liana.method._pipe_utils._common import _get_props

from liana.method.sp._misty._Misty import MistyData

from liana.utils.spatial_neighbors import spatial_neighbors
from liana.method._pipe_utils._pre import _choose_mtx_rep

from liana.resource import select_resource
from liana.method._pipe_utils import prep_check_adata
from liana.method.sp._utils import _add_complexes_to_var

def _make_view(adata, nz_threshold=0.1, add_obs=False, use_raw=False,
               layer=None, connecitivity=None, spatial_key=None, verbose=False):

    X = _choose_mtx_rep(adata=adata, use_raw=use_raw, layer=layer, verbose=verbose)

    obsm = dict()
    obsp = dict()
    if spatial_key is not None:
        if spatial_key not in adata.obsm.keys():
            raise ValueError(f"spatial_key {spatial_key} not found in `obsm`")
        obsm[spatial_key] = adata.obsm[spatial_key]

        if connecitivity is not None:
            obsp = dict()
            obsp[f"{spatial_key}_connectivities"] = connecitivity

    obs = adata.obs if add_obs else pd.DataFrame(index=adata.obs.index)

    adata = AnnData(X=X, obs=obs, var=pd.DataFrame(index=adata.var_names),
                    obsp=obsp, obsm=obsm, dtype=np.float32)
    var_msk = _get_props(adata.X) >= nz_threshold
    adata = adata[:, var_msk]

    return adata


def genericMistyData(intra,
                     intra_use_raw=False,
                     intra_layer=None,
                     extra=None,
                     extra_use_raw=False,
                     extra_layer=None,
                     nz_threshold=0.1,
                     add_para=True,
                     spatial_key='spatial',
                     set_diag=False,
                     kernel = 'misty_rbf', ## TODO change to gaussian kernel
                     bandwidth = 100,
                     zoi = 0,
                     cutoff = 0.1,
                     add_juxta=True,
                     n_neighs = 6,
                     verbose=False
                     ):

    """
    Construct a MistyData object from an AnnData object with views as presented in the manuscript.

    Parameters
    ----------
    intra : `anndata.AnnData`
        AnnData object with the intraview
    intra_use_raw : `bool`, optional (default: False)
        Whether to use the raw data of the intraview.
    intra_layer : `str`, optional (default: None)
        The layer of the intraview to use.
    extra : `anndata.AnnData`, optional (default: None)
        AnnData object with the extraview(s). If None, the extraview is set to be the same as the intraview.
    extra_use_raw : `bool`, optional (default: False)
        Whether to use the raw data of the extraview.
    extra_layer : `str`, optional (default: None)
        The layer of the extraview(s) to use.
    nz_threshold: `float`, optional (default: 0.1)
        The threshold for the number of non-zero entries in each view.
    add_para : `bool`, optional (default: True)
        Whether to add the paraview.
    spatial_key : `str`, optional (default: 'spatial')
        The key in adata.obsm where the spatial coordinates are stored.
    set_diag : `bool`, optional (default: True)
        Whether to set the diagonal of the connectivity matrix to 1.
    kernel : `str`, optional (default: 'misty_rbf')
        A radial basis function kernel to use for the generation of the connectivity matrix for the paraview.
        Default is 'misty_rbf', a kernel derivative of a Gaussian kernel.
    bandwidth : `float`, optional (default: 100)
        The bandwidth of the kernel.
    zoi : `float`, optional (default: 0)
        The zone of indifference of the kernel, i.e. the kernel is set to 0 for distances smaller than zoi.
    cutoff : `float`, optional (default: 0.1)
        The cutoff for the connectivity matrix.
    add_juxta : `bool`, optional (default: True)
        Whether to add the juxtaview. The juxtaview is constructed using only the nearest neighbors.
        A bandwidth of 5 times the bandwidth of the paraview is used to ensure that the nearest neighbors within the radius.
    n_neighs : `int`, optional (default: 6)
        The number of neighbors to consider when constructing the juxtaview.
    verbose : `bool`, optional (default: False)
        Whether to print progress.

    Returns
    -------
    `MistyData` object with the intra view, and two fixed extra view(s): para and juxta.

    """
    # init views
    views = {}
    intra = _make_view(adata=intra, nz_threshold=nz_threshold, add_obs=True,
                       use_raw=intra_use_raw, layer=intra_layer,
                       spatial_key=spatial_key, verbose=verbose)
    views['intra'] = intra

    if extra is None:
        extra = intra

    if add_juxta:
        neighbors = spatial_neighbors(extra,
                                      bandwidth=bandwidth*5,
                                      spatial_key=spatial_key,
                                      max_neighbours=n_neighs,
                                      set_diag=set_diag,
                                      inplace=False)

        views['juxta'] = _make_view(adata=extra, nz_threshold=nz_threshold,
                                    use_raw=extra_use_raw, layer=extra_layer,
                                    spatial_key=spatial_key, connecitivity=neighbors,
                                    verbose=verbose)

    if add_para:
        weights = spatial_neighbors(adata=extra,
                                    spatial_key=spatial_key,
                                    bandwidth=bandwidth,
                                    kernel=kernel,
                                    set_diag=set_diag,
                                    inplace=False,
                                    cutoff=cutoff,
                                    zoi=zoi
                                    )
        views['para'] = _make_view(adata=extra, nz_threshold=nz_threshold,
                                   use_raw=extra_use_raw, layer=extra_layer,
                                   spatial_key=spatial_key, connecitivity=weights,
                                   verbose=verbose)

    return MistyData(views, intra.obs, spatial_key)


def _check_if_squidpy() -> ModuleType:
    try:
        import squidpy as sq
    except Exception:

        raise ImportError(
            'squidpy is not installed. Please install it with: '
            'pip install squidpy'
        )
    return sq


def lrMistyData(adata,
                resource_name='consensus',
                resource=None,
                nz_threshold=0.1,
                use_raw = False,
                layer = None,
                spatial_key='spatial', ## TODO Change to Gaussian kernel
                kernel = 'misty_rbf',
                bandwidth = 100,
                set_diag = False,
                cutoff = 0.1,
                zoi = 0,
                verbose = False
                ):
    """
    Generate a MistyData object from an AnnData object in ligand-receptor format.

    Parameters
    ----------
    adata : `anndata.AnnData`
        AnnData object
    resource_name : `str`, optional (default: 'consensus')
        The name of the resource to use. See `show_resources` for available resources.
    resource : `pandas.DataFrame`, optional (default: None)
        A resource in the form of a pandas DataFrame. If None, the resource is selected using `select_resource`.
    nz_threshold : `float`, optional (default: 0.1)
        The threshold for the number of non-zero entries in each view.
    use_raw : `bool`, optional (default: False)
        Whether to use the raw data of the AnnData object.
    layer : `str`, optional (default: None)
        The layer of the AnnData object to use.
    spatial_key : `str`, optional (default: 'spatial')
        The key in adata.obsm where the spatial coordinates are stored.
    kernel : `str`, optional (default: 'misty_rbf')
        A radial basis function kernel to use for the generation of the connectivity matrix for the extra view.
        Default is 'misty_rbf', a kernel derivative of a Gaussian kernel.
    bandwidth : `float`, optional (default: 100)
        The bandwidth of the kernel.
    set_diag : `bool`, optional (default: True)
        Whether to set the diagonal of the connectivity matrix to 1.
    cutoff : `float`, optional (default: 0.1)
        The minimum value cutoff for the connectivity matrix.
    zoi : `float`, optional (default: 0)
        Zone of indifference of the kernel, i.e. the kernel is set to 0 for distances smaller than zoi.
    verbose : `bool`, optional (default: False)
        Whether to print progress.

    Returns
    -------
    A `MistyData` object with receptors in the intra view & ligands in the extra view.
    """
    # TODO: reduce redundancies in documentation
    if resource is None:
        resource = select_resource(resource_name)

    adata = prep_check_adata(adata=adata,
                             use_raw=use_raw,
                             layer=layer,
                             verbose=verbose,
                             groupby=None,
                             min_cells=None,
                             obsm = {spatial_key: adata.obsm[spatial_key]}
                             )

    adata = _add_complexes_to_var(adata,
                                  np.union1d(resource['receptor'].astype(str),
                                             resource['ligand'].astype(str)
                                             )
                                  )

    # filter_resource after adding complexes to var
    resource = resource[(np.isin(resource.ligand, adata.var_names)) &
                        (np.isin(resource.receptor, adata.var_names))]

    views = dict()
    views['intra'] =  _make_view(adata=adata[:, resource['receptor'].unique()],
                        nz_threshold=0, add_obs=True)

    connectivity = spatial_neighbors(adata=adata, spatial_key=spatial_key,
                                     bandwidth=bandwidth, kernel=kernel,
                                     set_diag=set_diag, cutoff=cutoff,
                                     zoi=zoi, inplace=False)

    views['extra'] = _make_view(adata=adata[:,resource['ligand'].unique()],
                                spatial_key=spatial_key, nz_threshold=nz_threshold,
                                connecitivity=connectivity)

    return MistyData(data=views, obs=views['intra'].obs, spatial_key=spatial_key)



import numpy as np
import pandas as pd

def _sample_dea(adata, groupby):
    nrow = adata.n_vars * 5

    rng = np.random.default_rng(1337)

    dea_df = pd.DataFrame({'gene': rng.choice(adata.var_names, nrow),
                            'stat': rng.random(nrow),
                            'pval': rng.random(nrow),
                            'padjusted': rng.random(nrow),
                            groupby: rng.choice(adata.obs[groupby].unique(), nrow)
                            })
    dea_df = dea_df.drop_duplicates(['gene', groupby]).set_index("gene")

    return dea_df
